#!/usr/bin/env python3
# you schould not need to do anything in here
import sys                           #needed for executing this sript in diverent modes
import subprocess as sp              #needed to run mcstas
import argparse
import os
from importlib import import_module
from mcpw.mcstas_wrapper import run_mcstas, run_compiler, run_instrument,\
                            valid_config, valid_mcconfig,\
                            psave, pload, load_var_list, save_var_list,\
                            check_for_detector_output, get_result_path_from_input
from os.path import isfile, isabs
import csv
#for exeption handling
class ArgumentParserError(Exception):
    pass
class ThrowingArgumentParser(argparse.ArgumentParser):
    def error(self, message):
        raise ArgumentParserError(message)


def main():
    # defining the argument parser for nice usage
    parser= ThrowingArgumentParser(description="Control command for automatic mcstas usage for the Reseda Instrument")

    subparsers = parser.add_subparsers(title="modes", required=True,
             description="Use 'manager mode --help' to view the help for any command.",
             metavar='mode')

    parser.add_argument('-p', '--python_file', dest='python_file', default='', type=str, required=True,
                       help='path (absolute or not) to the python file containing mcvariables and analyse functions')

    parser.add_argument('-d', '--result-dir', dest='result_dir', default='', type=str,
                       help='directory name for the simulationresults. If none is given or foulder name exists allrady, a increment Folder Name is generated')

    parser.add_argument('-l', '--list', dest='list', default='', type=str,
                       help='a file containing variables in csv format. first row has to match with variable names occuring in the instrument. variables that are not in here take the value from the py file. each row is a single simmulation. list will be saved together with the result for later use')

    parser_server = subparsers.add_parser('server', help='mcstas will be executed localy and the processed simulation results packed in a tarball')
    parser_server.set_defaults(func='server')

    parser_remote = subparsers.add_parser('remote', help='mcstas will be executed on a remote machine')
    parser_remote.set_defaults(func='remote')

    parser_local = subparsers.add_parser('local', help='mcstas will be executed on localy')
    parser_local.set_defaults(func='local')

    parser_full = subparsers.add_parser('full', help='mcstas will be executed on localy and the analyze function will be called after')
    parser_full.set_defaults(func='full')

    parser_analyse = subparsers.add_parser('analyse', help='analyse function will be called')
    parser_analyse.set_defaults(func='analyse')

    parser_analyse = subparsers.add_parser('custom', help='runs the mcstas compiler, c compiler and a everyting that is in the function called custom in your python file. analyse function will be called !!!EXTREMLY EXPERIMENTAL!!!')
    parser_analyse.set_defaults(func='custom')


    #retreiveng command arguments
    try:
        args = parser.parse_args()
    except ArgumentParserError as e:
        parser.print_help()
        print(f"\n{e}")
        exit(1)


    # importing var and mcvar from local_var['port']py and reseda.py files
    sys.path.append(os.getcwd())
    sys.path.append(os.path.dirname(args.python_file))
    from local_var import variables  as var
    #initializing and validating local vars
    #var = variables()
    valid_config(var)

    # adding local working directory to PATH
    sys.path.append(var['p_local'])

    # adding pyhton file location to PATH
    #importing python_file as module
    pyinstr = import_module(f"{os.path.basename(args.python_file).split('.')[0]}")
    # pulling relevant functions and classes from python file
    mcvariables = vars(pyinstr)['mcvariables']
    mcvar = vars(pyinstr)['mcvariables']
    analyse = vars(pyinstr)['analyse']
    post_simulation = vars(pyinstr)['post_simulation']
    pre_simulation = vars(pyinstr)['pre_simulation']
    # initializing mcvariables class
    #mcvar = mcvariables()
    var_list = []
    if args.list:
        if not isabs(args.list):
            if not isfile(var['p_local']/args.list):
                sys.exit(f"could not find {args.list} in {var['p_local']}")
        else:
            if not isfile(var['p_local']/args.list):
                sys.exit(f"could not find {args.list}")
        try:
            if isabs(args.list):
                var_list = load_var_list(args.list)
            else:
                var_list = load_var_list(var['p_local']/args.list)
        except Exception as e:
            sys.exit(f"could not import {args.list}. an exception occurred:\n{e}")

    msg  = ''
    #calling the correct function according to command arguments
    if 'func' not in args:
        # Fallback to --help.
        parser.print_help()
    else:
        (mcvar["sim"],msg) = get_result_path_from_input(var, mcvar, msg, args) # logic for retreiveng the correct name for the result foulder
        valid_mcconfig(var,mcvar)
        if args.func == 'analyse':
            try:
                var_list = load_var_list(var['sim_res']/mcvar["sim"]/'var_list')
            except: var_list = []
            mcvar = pload(var['sim_res']/mcvar["sim"]/'variables') #loading the correct variables
            valid_mcconfig(var,mcvar)
            check_for_detector_output(var,mcvar, var_list)
            analyse(var, mcvar, var_list)                             #call analyse, defined in reseda.py
            exit(0)
        elif args.func == 'server':#this runs the script in server mde
            run_mcstas(var,mcvar)
            run_compiler(var,mcvar)
            var,mcvar,var_list=pre_simulation(var,mcvar,var_list)
            run_instrument(var,mcvar, var_list)
            check_for_detector_output(var,mcvar,var_list)
            save_var_list(var_list, var['sim_res']/mcvar["sim"]/'var_list')  #save mcstas variables
            psave(mcvar, var['sim_res']/mcvar["sim"]/'variables')  #save mcstas variables
            post_simulation(var, mcvar, var_list) # contains functions that get executed after mcstas finished and can i.e. reformate the output
            sp.run(['tar', '-cf' '{}/{}.tar'.format(var['sim_res'], mcvar["sim"]), var['sim_res']/mcvar["sim"]]) #compress data

        elif args.func == 'remote':#use this if you want to run the simulation on a remote machine (setup has to be done beforhand)
            sp.run(['scp', '-r', '-P', str(var['port']), var['instr_file'], '{}:{}'.format(var['server'], var['p_server'])])#copy mcstas-instrument to remote
            sp.run(['scp', '-r', '-P', str(var['port']), 'manager.py', '{}:{}'.format(var['server'], var['p_server'])])#copy this file to remote
            sp.run(['scp', '-r', '-P', str(var['port']), 'reseda.py', '{}:{}'.format(var['server'], var['p_server'])])#copy this file to remote
            sp.run(['ssh' , '-p', str(var['port']), var['server'], 'cd {}; python {}manager.py server {}'.format(var['p_server'], var['p_server'],mcvar["sim"])])#run this file with server atribute remote
            sp.run(['scp', '-l', str(var['rate']), '-r', '-P', str(var['port']), '{}:{}.tar'.format(var['server'], var['p_server']/var['sim_res']/mcvar["sim"]), var['p_local']])#download data from remote
            sp.run(['tar', '-xf', '{}.tar'.format(mcvar["sim"])])#decompress data

        elif args.func == 'local':#use this to run the script localy
            run_mcstas(var,mcvar)
            run_compiler(var,mcvar)
            var,mcvar,var_list=pre_simulation(var,mcvar,var_list)
            run_instrument(var,mcvar,var_list)
            check_for_detector_output(var,mcvar,var_list)
            save_var_list(var_list, var['sim_res']/mcvar["sim"]/'var_list')  #save mcstas variables
            psave(mcvar, var['sim_res']/mcvar["sim"]/'variables')  #save mcstas variables
            post_simulation(var, mcvar,var_list) # contains functions that get executed after mcstas finished and can i.e. reformate the output

        elif args.func == 'full':
            run_mcstas(var,mcvar)
            run_compiler(var,mcvar)
            var,mcvar,var_list=pre_simulation(var,mcvar,var_list)
            run_instrument(var,mcvar,var_list)
            check_for_detector_output(var,mcvar,var_list)
            save_var_list(var_list, var['sim_res']/mcvar["sim"]/'var_list')  #save mcstas variables
            psave(mcvar, var['sim_res']/mcvar["sim"]/'variables')  #save mcstas variables
            post_simulation(var, mcvar, var_list) # contains functions that get executed after mcstas finished and can i.e. reformate the output
            analyse(var, mcvar,var_list)                                               #call analyse, defined in reseda.py

        elif args.func == 'custom':
            run_mcstas(var,mcvar)
            run_compiler(var,mcvar)
            custom = vars(pyinstr)['custom']
            custom(var,mcvar,var_list)
            analyse(var, mcvar)                                               #call analyse, defined in reseda.py
        print(msg)

if __name__ == '__main__':
    try:
        main()
    except Exception as e:
        print(f"an Error occured:\n\n{e}")



#!/usr/bin/env python3
# you schould not need to do anything in here
import sys                           #needed for executing this sript in diverent modes
import subprocess as sp              #needed to run mcstas
import argparse
import os
import re
from mcpw.setup_tools import create_local_var, create_python_file




#for exeption handling
class ArgumentParserError(Exception):
    pass
class ThrowingArgumentParser(argparse.ArgumentParser):
    def error(self, message):
        raise ArgumentParserError(message)


def main():
    # defining the argument parser for nice usage
    parser= ThrowingArgumentParser(description="Script to create a python file to control a McStas Instrument with MCPW")

    subparsers = parser.add_subparsers(title="modes",
             description="Use 'manager mode --help' to view the help for any command.",
             metavar='mode')

    parser.add_argument('-I', '--instrument', dest='instrument', default='', type=str, required=True,
                        help='mcstas instrument you want to control with mpw')
    parser.add_argument('-d', '--working_dir', dest='working_dir', default=os.getcwd(), type=str,
                        help='path to working directory (optional)')
    parser.add_argument('-m', '--mcstas', dest='mcstas', default='mcstas', type=str,
                        help='path to mcstas executable; default=mcstas')
    parser.add_argument('-o', '--output_dir', dest='output_dir', default='simulation_results', type=str,
                        help='name in the simulation results directory; default=simulation_results')
    parser.add_argument('-c', '--component_dir', dest='component_dir', default='', type=str,
                        help='additional directory where mcstas will search for components needs to be a absolute path; default=""')
    parser.add_argument('-M', '--MPI', dest='mpi', default=0, type=int,
                        help='additional directory where mcstas will search for components needs to be a absolute path; default=""')


    #retreiveng command arguments
    try:
        args = parser.parse_args()
    except ArgumentParserError as e:
        print(e)
        exit(1)

    if os.name == 'nt':
        args.working_dir   = re.sub(r'\\','/', args.working_dir)
        args.mcstas        = re.sub(r'\\','/', args.mcstas)
        args.instrument    = re.sub(r'\\','/', args.instrument)
        args.output_dir    = re.sub(r'\\','/', args.output_dir)
        args.component_dir = re.sub(r'\\','/', args.component_dir)


    if not os.path.isfile(f"{args.working_dir}/local_var.py"):
        create_local_var(args)
    else:
        print(f"local_var.py allready exists")

    if not os.path.isfile(f"{args.working_dir}/{args.instrument.split('.')[0]}.py"):
        create_python_file(args)
    else:
        print(f"{args.instrument.split('.')[0]}.py allready exists")


if __name__ == '__main__':
    main()





import subprocess as sp              #needed to run mcstas
import sys                           #needed to select program mode of this script
import pickle                        #needed to save mcstas variables for later use
import os
from os.path import isfile, isdir, isabs, dirname, basename, splitext, join, islink
import locale
from shutil import copyfile, which
import csv
from datetime import datetime

class DummyFile(object):
    def write(self, x): pass

class Scan():#helping class for easyer use of the scan funktionality of mcstas, needs start and stop value and the unit of the values as well the numbers of steps
    def __init__(self, start, stop, unit, N):
        self.start = start
        self.stop = stop
        self.range = stop-start
        self.unit = unit
        self.N = N
        self.mc = '{},{}'.format(start, stop)
        self.step = (stop-start) / (N-1)
    def absolute_value(self, n):
        return self.step * n + self.start

def scan(mcvar):
    for key,value in mcvar.items():
        if isinstance(value,Scan):
            return value
    print("no object of Class Scan found")
    return None

def execute(command, errormsg, successmsg, print_command=True):
    if print_command:
        print(f"runing: {command}")
    run_return = sp.run(command, shell=True, text=True, stdout=sp.PIPE, stderr=sp.PIPE)
    # check if the process was successfully
    if run_return.returncode != 0:
        print(run_return.returncode)
        print(run_return.stdout)
        print(run_return.stderr)
        sys.exit(errormsg)
    else:
        #print(run_return.stdout)
        print(f"{successmsg}\n")
    return run_return

def valid_config(var):
    #check dirs
    if not isdir(var['p_local']):
        sys.exit(f"the given local working directory '{var['p_local']}' dose not exist")
    if not isdir(var['p_server']):
        sys.exit(f"the given server working directory '{var['p_server']}' dose not exist\n\
                 if you dont intend to use the ssh feature, just set it so valid path")
    if isabs(var['sim_res']):
        if not isdir(var['sim_res']):
            sys.exit(f"the given simulation result directory '{var['sim_res']}' dose not exist")
    else:
        if not isdir(var['p_local']/var['sim_res']):
            os.mkdir(var['p_local']/var['sim_res'])
            print(f"created simulation result directory '{var['sim_res']}' in the local working directory")
    if isabs(var['sim_res']): #making shure var['sim_res'] is allways a absolute path
        pass
    else:
        var['sim_res'] = var['p_local']/var['sim_res']

    mcstas = which(var['mcstas'])
    if not mcstas:
        sys.exit(f"\nMcStas is not installed or the Path is inocrrect: {var['mcstas']}")
    if islink(mcstas):
        var['mcstas'] = os.readlink(mcstas)

def valid_mcconfig(var,mcvar):
    if not mcvar['instr_file'].split('.')[1] == "instr":
        sys.exit(f"\nthe given instrument file has not the correct ending:\
                 \nexpected: {mcvar['instr_file'].split('.')[0]}.instr\
                 \ngot: {mcvar['instr_file']}\
                 \nplease make shure to give a valid instrument file")

    if not isfile(var['p_local']/mcvar['instr_file']):
        sys.exit(f"\nthe instrument file '{mcvar['instr_file']}' dose not exist")

    if var['mpi'] == 0:
        # test if gcc exists
        if not which("gcc"):
            sys.exit(f"\ngcc is not installed. please install it")
    else:
        # test if mpicc exists
        if os.name=='nt':
            if not which("mpiexec"):
                sys.exit(f"\nmpiexec is not installed. please install Microsoft MPI form the microsoft website")
        else:
            if not which("mpicc"):
                sys.exit(f"\nmpicc is not installed. please install openmpi")


def encode_files_to_local_encoding(file_or_dir_list, output_dir):
    file_list = []
    for entry in file_or_dir_list:
        if entry != '':
            if isdir(entry):
                with os.scandir(entry) as it:
                    for e in it:
                        if e.name.endswith(('.comp','.instr','.h')) and e.is_file():
                            file_list.append(e.path)
            else:
                file_list.append(entry)
    for file in file_list:
        try:
            of = open(file, mode='r', encoding='utf-8')
            content = of.read()
            of.close()
            file_name = basename(file)
            nf = open(f"{output_dir}/{file_name}", mode='w', encoding=locale.getpreferredencoding())
            nf.write(content)
            nf.close()
        except UnicodeDecodeError:
            copyfile(file, output_dir/basename(file))
        except UnicodeEncodeError:
            print(f' sciping {basename(file)}')

# mcstas -t --verbose -o outputfile instrfile (-I componentdir)
def run_mcstas(var, mcvar):
    # create output name
    instr_c_file = mcvar['instr_file'].split('.')[0] + ".c"
    temp_dir = 'temp_encoding'
    if os.name=='nt':
        os.environ['MCSTAS'] = dirname(var['mcstas'])+'/../lib/'
        if not isdir(var['p_local']/temp_dir):
            os.mkdir(var['p_local']/temp_dir)
        encode_files_to_local_encoding([var['p_local']/mcvar['instr_file'], var['p_local'], var['componentdir']], var['p_local']/temp_dir)
        run_string = f"{var['mcstas']} -t --verbose -o {var['p_local']/instr_c_file} {basename(mcvar['instr_file'])} -I {var['p_local']/temp_dir} "
        # exectue the run_string and capture the output
        print(f"runing: {run_string}")
        run_return = sp.run(run_string, shell=True, text=True, stdout=sp.PIPE, stderr=sp.PIPE, cwd=var['p_local']/temp_dir)
        for file in os.listdir(f"{var['p_local']/temp_dir}"):
            os.remove(f"{var['p_local']/temp_dir/file}")
        os.rmdir(f"{var['p_local']/temp_dir}")
    else:
        # create run_string
        run_string = f"{var['mcstas']} -t --verbose -o {var['p_local']/instr_c_file} {var['p_local']/mcvar['instr_file']} -I {var['p_local']} "
        #if a extra Component Dir is given add the option to the compile command
        if var['componentdir'] != "":
            if isabs(var['componentdir']):
                run_string = run_string + f"-I {var['componentdir']} "
            else:
                run_string = run_string + f"-I {var['p_local']/var['componentdir']} "
        # exectue the run_string and capture the output
        print(f"runing: {run_string}")
        run_return = sp.run(run_string, shell=True, text=True, stdout=sp.PIPE, stderr=sp.PIPE)
    # check if the process was successfull
    if run_return.returncode != 0:
        print(run_return.stdout)
        print(run_return.stderr)
        sys.exit("An error occurred while running McStas Compiler")
    else:
        print(f"McStas compiler done\n")

# gcc (mpicc) -o p_local/reseda.out p_local/reseda.c -lm (-DUSE_MPI) -g -O2 -lm -std=c99
def run_compiler(var,mcvar, cflags=""):
    instr_c_file = mcvar['instr_file'].split('.')[0] + ".c"
    if os.name=='nt':
        instr_out_file = mcvar['instr_file'].split('.')[0] + ".exe"
    else:
        instr_out_file = mcvar['instr_file'].split('.')[0] + ".out"
    #check if mpi is enabled
    if var['mpi'] == 0:
        run_string = f"gcc "
    else:
        if os.name=='nt':
            run_string = f"{dirname(var['mcstas'])}/mpicc.bat -DUSE_MPI "
        else:
            run_string = f"mpicc -DUSE_MPI "
    run_string = run_string + f"-o {var['p_local']/instr_out_file} {var['p_local']/instr_c_file} -lm -g -O2 -std=c99 {cflags}"
    # exectue the run_string and capture the output
    execute(run_string, "An error occurred while running the C Compiler", "C compiler done")

# (mpirun -np 2) instr.out -n -d var=value
def run_instrument(var,mcvar,var_list):
    sys.path.append(f"{os.path.dirname(var['mcstas'])}/../tools/Python/mcrun")
    from mccode import McStasResult
    errormsg = "The Simmulation Failed"
    successmsg = "The simmulation compleated successfully"
    no_use_vars = ["scan", "n", "sim", "instr_file"]
    if os.name=='nt':
        instr_out_file = mcvar['instr_file'].split('.')[0] + ".exe"
    else:
        instr_out_file = mcvar['instr_file'].split('.')[0] + ".out"
    params = ''
    scan_var = []
    res_list = []
    #check if mpi is enabled
    if var['mpi'] == 0:
        run_string = f"{var['p_local']/instr_out_file} -n {str(mcvar['n'])} "
    else:
        if os.name=='nt':
            run_string = f"mpiexec -np {var['mpi']} {var['p_local']/instr_out_file} -n {str(mcvar['n'])} "
        else:
            run_string = f"mpirun --use-hwthread-cpus -np {var['mpi']} {var['p_local']/instr_out_file} -n {str(mcvar['n'])} "
    # parsing the parameters and checking if a scan is required
    for var_name, var_value in mcvar.items():
    #for var_name, var_value in mcvar.__dict__.items():
        if not (var_name in no_use_vars):
            if isinstance(var_value, Scan):
                scan_var = [var_name, var_value]
            else:
                params = params + f"{var_name}={var_value} "
    # scan or no scan
    if var_list:
        params = ''
        dets_vals = []
        #creating main result directory
        os.mkdir(var['sim_res']/mcvar['sim'])
        for name in var_list[0]:
            no_use_vars.append(name)
        for i in range(len(var_list)-1):
            value_list=[]
            print(f"step:{i+1}/{len(var_list)-1}")
            for j, name in enumerate(var_list[0]):
                params = params + f"{name}={var_list[i+1][j]} "
                value_list.append(str(var_list[i+1][j]))
            for var_name, var_value in mcvar.items():
            #for var_name, var_value in mcvar.__dict__.items():
                if not (var_name in no_use_vars):
                    if isinstance(var_value, Scan):
                        sys.exit("VALUE ERROR: You can not have a Scan and the list opiton at the same time.\n Exiting")
                    else:
                        params = params + f"{var_name}={var_value} "
            final_run_string = run_string + f"-d {str(var['sim_res']/mcvar['sim']/str(i))} {params} "

            out = execute(final_run_string, errormsg, successmsg, print_command=False)
            sys.stdout = DummyFile()
            dets=McStasResult(out.stdout).get_detectors()
            sys.stdout  =sys.__stdout__
            for det in dets:
                #print(f'det.intensity: {det.intensity}, det.error: {det.error}')
                value_list.append(str(det.intensity))
                value_list.append(str(det.error))
            res_list.append(var['sim_res']/mcvar['sim']/str(i))
            dets_vals.append(value_list)
        create_sim_file(dets, var, mcvar, var_list)
        create_dat_file(dets, var, mcvar, var_list, dets_vals)

    elif scan_var:
        #scan
        #creating main result directory
        os.mkdir(var['sim_res']/mcvar['sim'])
        dets_vals = []
        #scanning all points
        print(f"running: {run_string} {scan_var[0]}={scan_var[1].mc}\n")
        for i in range (scan_var[1].N):
            value_list=[str(scan_var[1].absolute_value(i))]
            print(f"step: {scan_var[0]}={scan_var[1].absolute_value(i)}")
            i_params = params + f"{scan_var[0]}={scan_var[1].absolute_value(i)} "
            final_run_string = run_string + f"-d {str(var['sim_res']/mcvar['sim']/str(i))} {i_params} "

            out = execute(final_run_string, errormsg, successmsg, print_command=False)
            sys.stdout = DummyFile()
            dets=McStasResult(out.stdout).get_detectors()
            sys.stdout  =sys.__stdout__
            for det in dets:
                value_list.append(str(det.intensity))
                value_list.append(str(det.error))
            res_list.append(var['sim_res']/mcvar['sim']/str(i))
            dets_vals.append(value_list)
        create_sim_file(dets, var, mcvar, var_list)
        create_dat_file(dets, var, mcvar, var_list, dets_vals)
    else:
        final_run_string = run_string + f"-d {str(var['sim_res']/mcvar['sim'])} {params} "
        execute(final_run_string, errormsg, successmsg)
        res_list.append(var['sim_res']/mcvar['sim'])
    return res_list

def psave(obj, file_path):#saves the given object as a pickle dump in the given file (file gets created)
    if obj:
        f = open(file_path, mode='xb')
        pickle.dump(obj, f)
        f.close

def pload(file_path):#funktion can read file writen by the psave function and returns its content
    f = open(file_path, mode='rb')
    obj = pickle.load(f)
    f.close
    return obj

def load_var_list(file_path):
    var_list = []
    with open(file_path, newline='') as csvfile:
        csvdata = csv.reader(csvfile)# delimiter=',', quotechar='|')
        for row in csvdata:
            var_list.append(row)
    for x in range(len(var_list)-1):
        for y in range(len(var_list[0])):
            var_list[x+1][y] = float(var_list[x+1][y])
    return var_list

def save_var_list(var_list, filename):
    with open(filename, mode='w') as csvfile:
        w = csv.writer(csvfile)
        for i in range(len(var_list)):
            w.writerow(var_list[i])

def check_scan(var, mcvar, msg): #ignore for now, is something i might implement later fully
    dir_num = len(os.listdir(var['sim_res']/mcvar['sim']))
    if dir_num-4 == scan(mcvar).N:
        return True
    else:
        msg = msg + f"the number of result folders dont correspont to the number of steps (#Dir:{dir_num} vs scan.N:{scan(mcvar).N})\n"
        return False

def scan_name(mcvar):
    for var_name, var_value in mcvar.items():
    #for var_name, var_value in mcvar.__dict__.items():
        if not var_name == "scan" and isinstance(var_value,Scan):
            return var_name
    return False

def check_for_detector_output(var, mcvar, var_list):
    if var_list:
        for i in range(len(var_list)-1):
            if not os.path.isdir(var['sim_res']/mcvar['sim']/str(i)):
                print(f"the mcstas output dir {mcvar['sim']}/{i} dose not exist.\nexiting")
                exit()
    elif scan_name(mcvar):
        for i in range(scan(mcvar).N):
            if not os.path.isdir(var['sim_res']/mcvar['sim']/str(i)):
                print(f"the mcstas output dir {mcvar['sim']}/{i} dose not exist.\nexiting")
                exit()
    else:
        if not os.path.isdir(var['sim_res']/mcvar['sim']):
            print(f"the mcstas output dir {mcvar['sim']} dose not exist.\n exiting")
            exit()
        else:
            #print("all fine")
            return

def get_result_path_from_input(var, mcvar, msg, args):# logic for retreiveng the correct name for the result foulder
    if args.func == 'analyse':
        if args.result_dir:
            return args.result_dir, msg
        else:
            d = var['sim_res']
            return sorted(d.iterdir(), key=os.path.getmtime, reverse=True)[0].name, msg
    if not args.result_dir:
        name = mcvar['sim']
    else:
        name = args.result_dir
    if os.path.isdir(var['sim_res']/name):
            counter = 0
            new_name = name + "_" + str(counter)
            while os.path.isdir(var['sim_res']/new_name):
                counter = counter + 1
                new_name = name + "_" + str(counter)
            msg = f'####################\n new result directory is {new_name}\n####################\n'
            return new_name, msg
    return name, msg

def mcplot(var,mcvar,msg='', mode='qt'):
    if mode == 'qt':
        run_string= f"{dirname(var['mcstas'])}/mcplot-pyqtgraph "
    else:
        run_string= f"{dirname(var['mcstas'])}/mcplot-matplotlib "
    run_string=run_string+f"{var['sim_res']/mcvar['sim']}"
    sp.run(run_string, shell=True)



def create_sim_file(dets, var, mcvar,var_list):
    if var_list:
        steps = len(var_list)-1
        scan_names = ", ".join(var_list[0])
        params = ""
        for i in range(len(var_list[0])):
            #print(f' {var_list[0][i]} = {var_list[1][i]}, {var_list[0][i]} = {var_list[-1][i]},')
            params+=f' {var_list[0][i]} = {var_list[1][i]}, {var_list[0][i]} = {var_list[-1][i]},'
        params = params[:-1] # removing last comma
        xlimits = f' {var_list[1][0]} {var_list[-1][0]}'
    else:
        steps = scan(mcvar).N
        scan_names = scan_name(mcvar)
        params = f" {scan_names} = {scan(mcvar).start}, {scan_names} = {scan(mcvar).stop}"
        xlimits = f" {scan(mcvar).start} {scan(mcvar).stop}"
    lines = []
    lines.append(f'begin instrument:')
    version = execute(f"{var['mcstas']} -v", "","",print_command=False).stdout.split(" ")[2]
    lines.append(f'  Creator: mcstas {version}')
    lines.append(f"  Source: {mcvar['instr_file']}")
    lines.append(f'  Parameters: {scan_names}')
    lines.append(f'  Trace_enabled: no')
    lines.append(f'  Default_main: yes')
    lines.append(f'  Embedded_runtime: yes')
    lines.append(f'end instrument:')
    lines.append(f'')
    lines.append(f'begin simulation')
    lines.append(f'Date: {datetime.strftime(datetime.now(),"%a %b %d %H %M %Y")}')
    lines.append(f"Ncount: {mcvar['n']}")
    lines.append(f'Numpoints: {steps}')
    lines.append(f'Param:{params}')
    lines.append(f'end simulation')
    lines.append(f'')
    lines.append(f'begin data')
    lines.append(f'type: multiarray_1d({steps})')
    lines.append(f'title: Scan of {scan_names}')
    lines.append(f'xvars: {scan_names}')
    det_string=""
    variables_string=""
    for det in dets:
        det_string+= f' ({det.name}_I,{det.name}_ERR)'
        variables_string+= f' {det.name}_I {det.name}_ERR'
    lines.append(f'yvars:{det_string}')
    lines.append(f"xlabel: '{scan_names}'")
    lines.append(f"ylabel: 'Intensity'")
    lines.append(f'xlimits:{xlimits}')
    lines.append(f'filename: mccode.dat')
    lines.append(f'variables: {scan_names.replace(",","")}{variables_string}')
    lines.append(f'end data')

    with open(var['sim_res']/mcvar['sim']/'mccode.sim', "w") as simfile:
        for line in lines:
            simfile.write("{}\n".format(line))

def create_dat_file(dets, var, mcvar, var_list, dets_vals):
    if var_list:
        steps = len(var_list)-1
        scan_names = ", ".join(var_list[0])
        params = ""
        for i in range(len(var_list[0])):
            params+=f' {var_list[0][i]} = {var_list[1][i]},'
        params = params[:-1] # removing last comma
        xlimits = f' {var_list[1][0]} {var_list[-1][0]}'
    else:
        steps = scan(mcvar).N
        scan_names = scan_name(mcvar)
        params = f" {scan_names} = {scan(mcvar).start}"
        xlimits = f" {scan(mcvar).start} {scan(mcvar).stop}"
    lines = []
    lines.append(f"# Instrument-source: '{mcvar['instr_file']}'")
    lines.append(f'# Date: {datetime.strftime(datetime.now(),"%a %b %d %H %M %Y")}')
    lines.append(f"# Ncount: {mcvar['n']}")
    lines.append(f'# Numpoints: {steps}')
    lines.append(f'# Param:{params}')
    lines.append(f'# type: multiarray_1d({steps})')
    lines.append(f'# title: Scan of {scan_names}')
    lines.append(f"# xlabel: '{scan_names}'")
    lines.append(f"# ylabel: 'Intensity'")
    lines.append(f'# xvars: {scan_names}')
    det_string=""
    variables_string=""
    for det in dets:
        det_string+= f' ({det.name}_I,{det.name}_ERR)'
        variables_string+= f' {det.name}_I {det.name}_ERR'
    lines.append(f'# yvars:{det_string}')
    lines.append(f'# xlimits:{xlimits}')
    lines.append(f'# filename: mccode.dat')
    lines.append(f'# variables: {scan_names.replace(",","")}{variables_string}')
    for row in dets_vals:
        lines.append(f'{" ".join(row)}')

    with open(var['sim_res']/mcvar['sim']/'mccode.dat', "w") as simfile:
        for line in lines:
            simfile.write("{}\n".format(line))



# you schould not need to do anything in here
import sys                           #needed for executing this sript in diverent modes
import subprocess as sp              #needed to run mcstas
import argparse
import os
import re
from os.path import isdir, isfile, isabs, islink
from mcpw.setup_tools import create_local_var, create_mcvar_dict
from mcpw.mcstas_wrapper import run_mcstas, run_compiler,\
                                run_instrument, is_scan,\
                                check_for_detector_output,\
                                psave, pload, which, valid_config,\
                                save_var_list, scan
from mcpw.mcstas_wrapper import load_var_list as mcpw_load_var_list
def load_mcvariables(var, sim=''): # loads used parameters from simulation
    if not sim:
        print('no result folder name given.\n please enter one as 2nd argument to this function.')
        return
    mcvar =  pload(var['sim_res']/sim/'variables') #loading the correct variables
    print(f'available parameters in mcvariables: {", ".join(list(mcvar.keys()))}')
    return mcvar

def load_var_list(var,mcvar):
    try:
        return mcpw_load_var_list(var['sim_res']/mcvar['sim']/'var_list')
    except:
        return []

def simulate(var, mcvar, var_list=[], var_list_csv='', sim='', remote=False): #spawns a simulation if sim dose not jet exists and returns a list of result dirs
    msg = ''
    #checking for imput
    if not sim:
        print('no result folder name given.\n please enter one as 4th argument to this function.')
        return

    # checking if the result folder exists and if yes, load mcvariables and create output list
    if os.path.isdir(var['sim_res']/sim):
        print('A Simulation with this result foder name allrady exists, skip Simulation.\n')
        res_list = []
        if is_scan(mcvar):
            for i in range(scan(mcvar).N):
                res_list.append(var['sim_res']/mcvar['sim']/str(i))
        else:
            res_list.append(var['sim_res']/mcvar['sim'])
        return load_mcvariables(var, sim), res_list
    else:
        mcvar['sim'] = sim

    # check if var_list or var_list file is given.
    if var_list_csv and not var_list:
        if not isabs(var_list_csv):
            if not isfile(var['p_local']/var_list_csv):
                sys.exit(f"could not find {var_list_csv} in {var['p_local']}")
        else:
            if not isfile(var['p_local']/var_list_csv):
                sys.exit(f"could not find {var_list_csv}")
        try:
            if isabs(var_list_csv):
                var_list = mcpw_load_var_list(var_list_csv)
            else:
                var_list = mcpw_load_var_list(var['p_local']/var_list_csv)
        except Exception as e:
            sys.exit(f"could not import {var_list_csv}. an exception occurred:\n{e}")

    # if remote simulation is True (setup has to be done beforhand)
    if remote:
        sp.run(['scp', '-r', '-P', str(var['port']), var.instr_file, '{}:{}'.format(var['server'], var['p_server'])])#copy mcstas-instrument to remote
        sp.run(['scp', '-r', '-P', str(var['port']), 'manager.py', '{}:{}'.format(var['server'], var['p_server'])])#copy this file to remote
        sp.run(['scp', '-r', '-P', str(var['port']), 'reseda.py', '{}:{}'.format(var['server'], var['p_server'])])#copy this file to remote
        sp.run(['ssh' , '-p', str(var['port']), var['server'], 'cd {}; python {}manager.py server {}'.format(var['p_server'], var['p_server'],mcvar['sim'])])#run this file with server atribute remote
        sp.run(['scp', '-l', str(var['rate']), '-r', '-P', str(var['port']), '{}:{}.tar'.format(var['server'], var.p/var['p_server']/var['sim_res']/mcvar['sim']), var['p_local']])#download data from remote
        sp.run(['tar', '-xf', '{}.tar'.format(mcvar['sim'])])#decompress data
        print('remote simulation successfully\n')
        res_list = []
        if is_scan(mcvar):
            for i in range(scan(mcvar).N):
                res_list.append(var['sim_res']/mcvar['sim']/str(i))
        else:
            res_list.append(var['sim_res']/mcvar['sim'])
        return mcvar, res_list

    # run the script localy
    else:
        run_mcstas(var,mcvar)
        run_compiler(var,mcvar)
        res = run_instrument(var,mcvar, var_list)
        check_for_detector_output(var,mcvar, var_list)
        save_var_list(var_list, var['sim_res']/mcvar['sim']/'var_list')  #save mcstas variables
        psave(mcvar, var['sim_res']/mcvar['sim']/'variables')  #save mcstas variables
        print('simulation successfully\n')
        return mcvar, res

# helper function
def print_mcvariable_from_instrument(instrument):
    for line in create_class_mcvariables_lines(instrument):
        print(line)

# fuction to create local_var.py and load it
def initialize(instrument='', working_dir=os.getcwd(), mcstas='mcstas', output_dir='simulation_results', component_dir='', mpi=0):
    # substituting \ for / to avoid complications with the paths
    if os.name == 'nt':
        working_dir   = re.sub(r'\\','/', working_dir)
        mcstas        = re.sub(r'\\','/', mcstas)
        instrument    = re.sub(r'\\','/', instrument)
        output_dir    = re.sub(r'\\','/', output_dir)
        component_dir = re.sub(r'\\','/', component_dir)

    # checking instrument file
    instr_file = working_dir + "/" + instrument
    if not isfile(instr_file):
        sys.exit(f" the instrument file '{instrument}' is not located in the working directory '{working_dir}'")
    if not instrument.split('.')[1] == "instr":
        sys.exit(f"the given instrument file has not the correct ending:\
                 \nexpected: {instrument.split('.')[0]}.instr\
                 \ngot: {instrument}\
                 \nplease make shure to give a valid instrument file")

    # checking working directory
    if not isdir(working_dir):
        sys.exit(f"the given working directory '{working_dir}' dose not exist")

    # checking output directory
    if isabs(output_dir):
        if not isdir(output_dir):
            sys.exit(f"the given simulation result directory '{output_dir}' dose not exist")
    else:
        if not isdir(f"{working_dir}/{output_dir}"):
            os.mkdir(f"{working_dir}/{output_dir}")
            print(f"created simulation result directory '{output_dir}' in the working directory")

    # checking for mcstas executable
    mcstas_exe = which(mcstas)
    if not mcstas_exe:
        print(f"\nMcStas is not installed or the Path is inocrrect: {mcstas}")
        return
    if islink(mcstas_exe):
        mcstas = os.readlink(mcstas_exe)
    else:
        mcstas = mcstas_exe

    # checking gcc or mpi existence
    if mpi == 0:
        # test if gcc exists
        if os.name=='nt':
            which("gcc --help")
        else:
            which("gcc")
    else:
        # test if mpicc exists
        if os.name=='nt':
            pass
        else:
            which("mpicc")

    # parsing arguments
    args= argparse.Namespace(working_dir=working_dir,\
                            mcstas=mcstas,\
                            instrument=instrument,\
                            output_dir=output_dir,\
                            component_dir=component_dir,\
                            mpi = mpi)

    # creating local_var.py file if not exiisting
    #TODO: check if imput correspont to existing local_var.py and give user feedback
    if not os.path.isfile(f"{working_dir}/local_var.py"):
        create_local_var(args)
    # importing local variables
    sys.path.append(working_dir)
    from local_var import variables
    var = variables()
    valid_config(var)
    print("usefull variables: var['sim_res'] directory where simulation_results are saved.\n\
                        in combination with mcvar['sim'] you get the full path to your detector files:\n\
                        var['sim_res']/mcvar['sim'] \n\
                        if you run a scan append /str(n) for n_th step\n\
                   var['p_local'] working directory\n")
    # printing mcvariables class
    print(f'####################\ncopy the following part and execute it in a new cell\n####################\n')
    print("from mcpw.mcstas_wrapper import Scan\n")
    for line in create_mcvar_dict(instr_file):
        print(line)
    print("\n#for laziness:")
    print("mcvar = mcvariables\n")
    print(f'####################\nend of cell\n####################')

    return var
